import typing as t
from datetime import datetime, timedelta
from enum import Enum
from io import BytesIO
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec.structs import asdict
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.recruitee.schemas import (
    RecruiteeJob,
    RecruiteeProfile,
)
from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)


class Sort(str, Enum):
    BY_DATE = "by_date"
    BY_LAST_MESSAGE = "by_last_message"


class RecruiteeEndpoint(str, Enum):
    DEVELOPMENT_ENDPOINT = "https://api.rc.recruitee.dev/c"
    STAGING_ENDPOINT = "https://api.s.recruitee.com/c"
    PRODUCTION_ENDPOINT = "https://api.recruitee.com/c"


class AuthParameters(Struct):
    company_id: Annotated[
        str,
        Meta(
            description="Company ID. A company subdomain can also be used.",
        ),
    ]
    api_token: Annotated[
        str,
        Meta(
            description=(
                "Personal API Token allowing access to the Recruitee API from external"
                " services."
            ),
        ),
    ]
    recruitee_endpoint: Annotated[
        RecruiteeEndpoint,
        Meta(
            description="Specifies which endpoint to be used, satging or production.",
        ),
    ] = RecruiteeEndpoint.PRODUCTION_ENDPOINT


class ReadProfilesParameters(Struct):
    limit: Annotated[
        t.Optional[int],
        Meta(
            description="Specifies the number of candidates to retrieve",
        ),
    ] = None
    offset: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "Skip number of candidates from the begining, used for ‘load more’,"
                " offset for next page should be current offset + limit"
            ),
        ),
    ] = None
    created_after: Annotated[
        t.Optional[str],
        Meta(
            description="Show only candidates created after given date",
        ),
    ] = None
    disqualified: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                """Show only disqualified candidates who are disqualified in at least"""
                """ one job (should be string ‘true’ or ‘1’)."""
            ),
        ),
    ] = None
    qualified: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Show only disqualified candidates who are qualified in at least one"
                " job (should be string ‘true’ or ‘1’)."
            ),
        ),
    ] = None
    ids: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "List of IDs separated by comma, example: 234221,4211412,535432"
            ),
        ),
    ] = None
    offer_id: Annotated[
        t.Optional[str],
        Meta(
            description="Filter by offer",
        ),
    ] = None
    query: Annotated[
        t.Optional[str],
        Meta(
            description="Search query for candidate’s name or offer",
        ),
    ] = None
    sort: Annotated[
        t.Optional[Sort],
        Meta(
            description="Sorting options: by_date, by_last_message",
        ),
    ] = None
    with_messages: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Show only candidates with messages (should be string ‘true’ or ‘1’)"
            ),
        ),
    ] = None
    with_my_messages: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Show only candidates with messages that current admin sent (should be"
                " string ‘true’ or ‘1’"
            ),
        ),
    ] = None


class WriteProfilesParameters(Struct):
    offers: Annotated[
        t.Optional[t.List[int]],
        Meta(
            description=(
                "Offers to which the candidate will be assigned with default stage."
            ),
        ),
    ] = None


class UpdateProfilesParameters(Struct):
    pass


class ArchiveProfilesParameters(Struct):
    pass


class Kind(str, Enum):
    JOB = "job"
    TALENT_POOL = "talent_pool"


class Scope(str, Enum):
    ARCHIVED = "archived"
    ACTIVE = "active"
    NOT_ARCHIVED = "not_archived"


class View_mode(str, Enum):
    DEFAULT = "default"
    BRIEF = "brief"


class ReadJobsParameters(Struct):
    kind: Annotated[
        t.Optional[Kind],
        Meta(
            description=(
                "If no kind is given, returns all job offers, if kind is job then lists"
                " only jobs, if scope is talent_pool, lists only talent pools"
            ),
        ),
    ] = Kind.JOB
    scope: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "If no scope is given list all job offers. archived returns only"
                " archived job offers, active returns published, internal and closed"
                " job offers, not_archived returns all but archived jobs"
            ),
        ),
    ] = Scope.ACTIVE
    view_mode: Annotated[
        View_mode,
        Meta(
            description=(
                "default (default mode, includes most of offer details); brief (only"
                " offer’s id, title, status and kind)"
            ),
        ),
    ] = View_mode.DEFAULT


def generic_candidates_read(
    mode: Mode,
):
    def read_candidates(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadProfilesParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        RECRUITEE_ENDPOINT = auth_parameters.recruitee_endpoint
        params = asdict(parameters)

        response = requests.get(
            "{}/{}/candidates".format(RECRUITEE_ENDPOINT, auth_parameters.company_id),
            headers={
                "Authorization": "Bearer {}".format(auth_parameters.api_token),
            },
            params=params,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull candidates list from Recruitee params={}"
                " status_code={} response={}".format(
                    params, response.status_code, response.text
                )
            )
        all_candidates = response.json()["candidates"]
        for candidate in all_candidates:
            if mode == Mode.create and not is_within_five_minutes(
                candidate["created_at"], candidate["updated_at"]
            ):
                continue
            if mode == Mode.update and is_within_five_minutes(
                candidate["created_at"], candidate["updated_at"]
            ):
                continue
            full_candidate_response = requests.get(
                "{}/{}/candidates/{}".format(
                    RECRUITEE_ENDPOINT, auth_parameters.company_id, candidate["id"]
                ),
                headers={
                    "Authorization": "Bearer {}".format(auth_parameters.api_token),
                },
            )
            if full_candidate_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to pull candidate details from Recruitee candidate_id={}"
                    " status_code={} response={}".format(
                        candidate["id"],
                        full_candidate_response.status_code,
                        full_candidate_response.text,
                    )
                )

            candidate = full_candidate_response.json()["candidate"]
            resume_url = candidate.get("cv_original_url")
            if resume_url:
                response = requests.get(url=resume_url)
                file = response.content
                profile_file = BytesIO(file)
                candidate["cv_file"] = profile_file
            else:
                candidate["cv_file"] = None

            yield candidate

    return read_candidates


def write_candidates(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[dict],
) -> t.List[t.Dict]:
    RECRUITEE_ENDPOINT = auth_parameters.recruitee_endpoint
    failed_profiles = []

    for profile in items:
        payload = {"candidate": profile, "offers": parameters.offers}

        response = requests.post(
            "{}/{}/candidates".format(RECRUITEE_ENDPOINT, auth_parameters.company_id),
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
                "Authorization": "Bearer {}".format(auth_parameters.api_token),
            },
            json=payload,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to add candidate to Recruitee status_code={} response={}"
                .format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


def update_candidates(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateProfilesParameters,
    items: t.Iterable[dict],
) -> t.List[t.Dict]:
    RECRUITEE_ENDPOINT = auth_parameters.recruitee_endpoint
    failed_profiles = []

    for profile in items:
        candidate_id = profile.pop("id")
        payload = {"candidate": profile}

        response = requests.patch(
            "{}/{}/candidates/{}".format(
                RECRUITEE_ENDPOINT, auth_parameters.company_id, candidate_id
            ),
            headers={
                "Content-Type": "application/json",
                "Authorization": "Bearer {}".format(auth_parameters.api_token),
            },
            json=payload,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to update candidate with id={} in Recruitee status_code={}"
                " response={}".format(
                    candidate_id,
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)

    return failed_profiles


def archive_candidates(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ArchiveProfilesParameters,
    items: t.Iterable[dict],
) -> t.List[t.Dict]:
    RECRUITEE_ENDPOINT = auth_parameters.recruitee_endpoint
    failed_profiles = []

    for profile in items:
        response = requests.delete(
            "{}/{}/candidates/{}".format(
                RECRUITEE_ENDPOINT, auth_parameters.company_id, profile["id"]
            ),
            headers={
                "Authorization": "Bearer {}".format(auth_parameters.api_token),
            },
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to archive candidate with id={} in Recruitee in Recruitee"
                " status_code={} response={}".format(
                    profile["id"],
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)

    return failed_profiles


def generic_jobs_read(
    mode: Mode,
):
    def read_jobs(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadJobsParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        RECRUITEE_ENDPOINT = auth_parameters.recruitee_endpoint
        params = asdict(parameters)

        response = requests.get(
            "{}/{}/offers".format(RECRUITEE_ENDPOINT, auth_parameters.company_id),
            headers={"Authorization": "Bearer {}".format(auth_parameters.api_token)},
            params=params,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Recruitee params={}"
                " status_code={} response={}".format(
                    params, response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from Recruitee")
        jobs = response.json()["offers"]
        for job in jobs:
            if mode == Mode.create and not is_within_five_minutes(
                job["created_at"], job["updated_at"]
            ):
                continue
            if mode == Mode.update and is_within_five_minutes(
                job["created_at"], job["updated_at"]
            ):
                continue
            full_job_response = requests.get(
                "{}/{}/offers/{}".format(
                    RECRUITEE_ENDPOINT, auth_parameters.company_id, job["id"]
                ),
                headers={
                    "Authorization": "Bearer {}".format(auth_parameters.api_token)
                },
            )
            if full_job_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to pull jobs details from Recruitee job_id={}"
                    " status_code={} response={}".format(
                        job["id"],
                        full_job_response.status_code,
                        full_job_response.text,
                    )
                )
                raise Exception("Failed to pull jobs from Recruitee")
            yield full_job_response.json()["offer"]

    return read_jobs


# To account for the time differnece between the start of job creation on the platform
# and its completion
def is_within_five_minutes(date_str1: str, date_str2: str) -> bool:
    date1 = datetime.fromisoformat(date_str1.replace("Z", "+00:00"))
    date2 = datetime.fromisoformat(date_str2.replace("Z", "+00:00"))

    difference = abs(date1 - date2)

    return difference <= timedelta(minutes=5)


ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=RecruiteeProfile,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=generic_candidates_read(Mode.create),
            update=generic_candidates_read(Mode.update),
            archive=generic_candidates_read(Mode.archive),
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=UpdateProfilesParameters,
            archive=ArchiveProfilesParameters,
        ),
        function=merge(
            create=write_candidates,
            update=update_candidates,
            archive=archive_candidates,
        ),
    ),
)


JobsAisle = Aisle(
    name=Entity.job,
    schema=RecruiteeJob,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(
            create=generic_jobs_read(Mode.create),
            update=generic_jobs_read(Mode.update),
            archive=generic_jobs_read(Mode.archive),
        ),
    ),
)



from hrflow_connectors.v2.connectors.flatchr.aisles import (
    AuthParameters,
    JobsAisle,
    ProfilesAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

FlatchrWarehouse = Warehouse(
    auth=AuthParameters,
    aisles=(JobsAisle, ProfilesAisle),
)



import typing as t

from hrflow_connectors.v2.connectors.flatchr.warehouse import FlatchrWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def format_flatchr_profile(flatchr_profile: t.Dict) -> t.Dict:
    hrflow_profile = dict(
        reference=flatchr_profile.get("id"),
        created_at=flatchr_profile.get("created_at"),
        info=dict(
            email=flatchr_profile.get("email"),
            first_name=flatchr_profile.get("firstname"),
            last_name=flatchr_profile.get("lastname"),
            full_name=(
                f"{flatchr_profile.get('firstname')} {flatchr_profile.get('lastname')}"
            ),
            location=dict(text="", lat=None, lng=None),
            phone=flatchr_profile.get("phone"),
        ),
        educations=[],
        experiences=[],
        attachments=[],
        tags=[
            {"name": "applicant_id", "value": flatchr_profile.get("applicant")},
            {"name": "vacancy", "value": flatchr_profile.get("vacancy")},
            {"name": "vacancy_id", "value": flatchr_profile.get("vacancy_id")},
            {"name": "source", "value": flatchr_profile.get("source")},
            {"name": "external_id", "value": flatchr_profile.get("external_id")},
            {"name": "hired", "value": flatchr_profile.get("hired")},
            {"name": "column", "value": flatchr_profile.get("column")},
        ],
        resume=flatchr_profile.get("resume"),
    )
    return hrflow_profile


def format_urls(urls: t.List) -> t.List:
    return [{f"{element['type']}": element["url"]} for element in urls]


def format_hrflow_profile(hrflow_profile: t.Dict) -> t.Dict:
    profile = dict(
        firstname=hrflow_profile["info"]["first_name"],
        lastname=hrflow_profile["info"]["last_name"],
        email=hrflow_profile["info"]["email"],
        phone=hrflow_profile["info"]["phone"],
        urls=format_urls(hrflow_profile["info"]["urls"]),
        type="link",
        resume=next(
            (
                attachment
                for attachment in hrflow_profile["attachments"]
                if attachment.get("type") == "resume"
            ),
            {},
        ).get("public_url"),
        source=f"HrFlow Source: {hrflow_profile['source']['name']}",
    )
    return profile


def format_hrflow_profile_for_update(hrflow_profile: t.Dict) -> t.Dict:
    flatchr_profile = format_hrflow_profile(hrflow_profile)
    flatchr_profile["id"] = hrflow_profile["reference"]
    return flatchr_profile


def format_flatchr_profile_for_archive(flatchr_profile: t.Dict) -> t.Dict:
    return {"reference": flatchr_profile["id"]}


def format_hrflow_profile_for_archive(hrflow_profile: t.Dict) -> t.Dict:
    vacancy_id = next(
        (tag["value"] for tag in hrflow_profile["tags"] if tag["name"] == "vacancy_id"),
        None,
    )
    applicant_id = next(
        (
            tag["value"]
            for tag in hrflow_profile["tags"]
            if tag["name"] == "applicant_id"
        ),
        None,
    )
    return {"applicant_id": applicant_id, "vacancy_id": vacancy_id}


def get_tags(flatchr_job: t.Dict) -> t.List:
    t = lambda name, value: dict(name=name, value=value)
    tags = [
        t("slug", flatchr_job.get("slug")),
        t("status", flatchr_job.get("status")),
        t("language", flatchr_job.get("language")),
        t("company_id", flatchr_job.get("company_id")),
        t("currency", flatchr_job.get("currency")),
        t("mensulaity", flatchr_job.get("mensulaity")),
        t("start_date", flatchr_job.get("start_date")),
        t("end_date", flatchr_job.get("end_date")),
        t("driver_license", flatchr_job.get("driver_license")),
        t("remote", flatchr_job.get("remote")),
        t("handicap", flatchr_job.get("handicap")),
        t("partial", flatchr_job.get("partial")),
    ]
    return tags


def format_flatchr_job(flatchr_job: t.Dict) -> t.Dict:
    hrflow_job = dict(
        reference=flatchr_job.get("id"),
        name=flatchr_job.get("title"),
        created_at=flatchr_job.get("created_at"),
        updated_at=flatchr_job.get("updated_at"),
        summary=flatchr_job.get("description"),
        location=dict(
            text="",
            lat=None,
            lng=None,
        ),
        sections=[
            dict(
                name="description",
                title="Description",
                description=flatchr_job.get("description"),
            ),
            dict(
                name="mission", title="Mission", description=flatchr_job.get("mission")
            ),
            dict(
                name="profile", title="Profile", description=flatchr_job.get("profile")
            ),
        ],
        responsibilities=flatchr_job.get("mission"),
        requirements=flatchr_job.get("profile"),
        url=flatchr_job.get("apply_url"),
        skills=[
            dict(name=skill, value=None, type="soft")
            for skill in (flatchr_job.get("skills") or "").split(";")
        ],
        tags=get_tags(flatchr_job),
    )
    return hrflow_job


def format_flatchr_job_for_archive(flatchr_job: t.Dict) -> t.Dict:
    return {"reference": flatchr_job["id"]}


DESCRIPTION = (
    "Avec le digital, recruter est devenu beaucoup plus compliqué qu'avant."
    " Flatchr centralise et simplifie vos processus de recrutement pour le rendre aussi"
    " simple que télécharger un film sur Netflix."
)
Flatchr = Connector(
    name="Flatchr",
    type=ConnectorType.ATS,
    subtype="flatchr",
    description=DESCRIPTION,
    url="https://www.flatchr.io/",
    warehouse=FlatchrWarehouse,
    flows=(
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_flatchr_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_flatchr_profile,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_flatchr_profile_for_archive,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_update,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_archive,
        ),
        Flow(
            Mode.create,
            Entity.job,
            Direction.inbound,
            format=format_flatchr_job,
        ),
        Flow(
            Mode.update,
            Entity.job,
            Direction.inbound,
            format=format_flatchr_job,
        ),
        Flow(
            Mode.archive,
            Entity.job,
            Direction.inbound,
            format=format_flatchr_job_for_archive,
        ),
    ),
)



import typing as t
from logging import LoggerAdapter
from typing import List

import requests
from msgspec import Meta, Struct
from msgspec.structs import asdict
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.hubspot.schemas import ContactObject
from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)

BASE_URL = "https://api.hubapi.com/crm/v3"
CONTACTS_ENDPOINT = "{}/objects/contacts".format(BASE_URL)


class AuthParameters(Struct):
    access_token: Annotated[
        str,
        Meta(
            description=(
                "The token used to authenticate any API calls made for to your HubSpot"
                " account."
            ),
        ),
    ]


class ReadProfilesParameters(Struct):
    properties: Annotated[
        str,
        Meta(
            description=(
                "A comma separated list of the properties to be returned in the"
                " response. If any of the specified properties are not present on the"
                " requested object(s), they will be ignored."
            ),
        ),
    ] = (
        "firstname,lastname,date_of_birth,email,phone"
        ",company,address,zip,city,state,country"
    )
    propertiesWithHistory: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "A comma separated list of the properties to be returned along with"
                " their history of previous values. If any of the specified properties"
                " are not present on the requested object(s), they will be ignored."
                " Usage of this parameter will reduce the maximum number of objects"
                " that can be read by a single request."
            ),
        ),
    ] = None
    associations: Annotated[
        t.Optional[List[str]],
        Meta(
            description=(
                "A comma separated list of object types to retrieve associated IDs for."
                " If any of the specified associations do not exist, they will be"
                " ignored."
            ),
        ),
    ] = None
    archived: Annotated[
        t.Optional[bool],
        Meta(
            description="Whether to return only results that have been archived.",
        ),
    ] = False


class Stage(Struct):
    label: str
    displayOrder: int
    metadata: str


class Pipeline(Struct):
    id: int
    label: str
    displayOrder: int
    stages: t.List[Stage]


class WriteProfilesParameters(Struct):
    dealID: t.Optional[int] = None


class UpdateProfilesParameters(Struct):
    objectWriteTraceId: t.Optional[str] = None


class ArchiveProfilesParameters(Struct):
    pass


def generic_read(mode: Mode):
    def read(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadProfilesParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        headers = {"Authorization": f"Bearer {auth_parameters.access_token}"}
        params = asdict(parameters)

        def fetch_contacts(params):
            response = requests.get(
                CONTACTS_ENDPOINT,
                headers=headers,
                params=params,
            )
            if response.status_code // 100 != 2:
                adapter.error(
                    f"Failed to pull contacts list from Hubspot. params={params} "
                    f"status_code={response.status_code} response={response.text}"
                )
                raise Exception("Failed to pull contacts list from Hubspot")
            return response.json()

        def should_yield_contact(contact):
            if (
                mode == Mode.create
                and contact["createdAt"][:16] != contact["updatedAt"][:16]
            ):
                return False
            if (
                mode == Mode.update
                and contact["createdAt"][:16] == contact["updatedAt"][:16]
            ):
                return False
            return True

        response_data = fetch_contacts(params)
        contacts = response_data["results"]

        for contact in contacts:
            if should_yield_contact(contact):
                yield contact

        while "paging" in response_data and "next" in response_data["paging"]:
            params["after"] = response_data["paging"]["next"]["after"]
            response_data = fetch_contacts(params)
            contacts = response_data["results"]
            for contact in contacts:
                if should_yield_contact(contact):
                    yield contact

    return read


def write(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    for profile in items:
        response = requests.post(
            CONTACTS_ENDPOINT,
            headers={
                "Authorization": "Bearer {}".format(auth_parameters.access_token),
            },
            json=profile,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to add profile to Hubspot status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
        else:
            contactId = response.json()["id"]
            adapter.info("Successfully added contact with ID={}".format(contactId))
            dealId = parameters.dealID
            if dealId is not None:
                deals_endpoint = (
                    "{}/objects/deals/{}/associations/contacts/{}/3".format(
                        BASE_URL, dealId, contactId
                    )
                )
                headers = {
                    "Authorization": "Bearer {}".format(auth_parameters.access_token),
                }
                deal_response = requests.request(
                    "PUT", deals_endpoint, headers=headers, data={}
                )
                if deal_response.status_code // 100 != 2:
                    adapter.error(
                        "Failed to add profile to Hubspot status_code={} response={}"
                        .format(
                            deal_response.status_code,
                            deal_response.text,
                        )
                    )
    return failed_profiles


def update(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    for profile in items:
        contactId = profile.pop("contactId")
        response = requests.patch(
            "{}/{}".format(CONTACTS_ENDPOINT, contactId),
            headers={
                "Authorization": "Bearer {}".format(auth_parameters.access_token),
            },
            json=profile,
            params=asdict(parameters),
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to update profile to Hubspot status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
        else:
            adapter.info("Successfully updated contact with ID={}".format(contactId))

    return failed_profiles


def archive(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ArchiveProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    for profile in items:
        contactId = profile["contactId"]
        response = requests.delete(
            "{}/{}".format(CONTACTS_ENDPOINT, contactId),
            headers={
                "Authorization": "Bearer {}".format(auth_parameters.access_token),
            },
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to archive profile to Hubspot status_code={} response={}"
                .format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
        else:
            adapter.info("Successfully archived contact with ID={}".format(contactId))
    return failed_profiles


ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=ContactObject,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=generic_read(Mode.create),
            update=generic_read(Mode.update),
            archive=generic_read(Mode.archive),
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=UpdateProfilesParameters,
            archive=ArchiveProfilesParameters,
        ),
        function=merge(create=write, update=update, archive=archive),
    ),
)



import datetime
import decimal

from msgspec import Struct


class AdmenMission(Struct):
    ID_MISSION: int
    ID_SOCIETE: int
    ID_CONSULTANT: int
    REF_INTERNE: str
    LIBELLE: str
    NB_RECRUT: int
    DESC_MISSION: str
    DATE_PROPOSITION: datetime.date
    MONTANT: decimal.Decimal
    ETAT: int
    TYPE_MISSION: int
    DATE_ACCORD_CLIENT: datetime.date
    DATE_PRESENTATION: datetime.date
    DATE_SIGNATURE: datetime.date
    DATE_ENTREE_FONCT: datetime.date
    SALAIRE_ENTREE: decimal.Decimal
    COMMENTAIRE: str
    FRAIS: float
    REGLE_FACT: int
    MONNAIE_FACT: str
    ANGLAIS: int
    CONDITION_REGL: str
    ABOUTIE: int
    LIBRE1: str
    LIBRE2: str
    LIBRE3: str
    LIBRE4: str
    ID_CONSULTANT_2: int
    FACT1: decimal.Decimal
    FACT2: decimal.Decimal
    FACT3: decimal.Decimal
    FACT4: decimal.Decimal
    FACT5: decimal.Decimal
    FACT6: decimal.Decimal
    DATE_FACT1: datetime.date
    DATE_FACT2: datetime.date
    DATE_FACT3: datetime.date
    DATE_FACT4: datetime.date
    DATE_FACT5: datetime.date
    DATE_FACT6: datetime.date
    ID_PERS1: int
    ID_PERS2: int
    ID_PERS3: int
    ID_PERS4: int
    SALAIRE_DEMANDE: decimal.Decimal
    SECTEUR: str
    FONCTION: str
    REMPLACEMENT: int
    DATE_DEB_RECHERCHE: datetime.date
    INTERROMPUE: int
    DATE_INTERROMPUE: datetime.date
    COMMENT_INTERROMPUE: str
    REPRISE: int
    DATE_REPRISE: datetime.date
    COMMENT_REPRISE: str
    PUBLIEE: int
    TYPECONTRAT: str
    REGION: str
    PAYS: str
    CITY: str
    SECTEURWEB: str
    FONCTIONWEB: str
    HTML: str
    SOFT_DELETED: int
    DATE_DELETED: datetime.date
    ID_PERDEL: int
    DATE_SAISIE: datetime.date
    COLOR: int
    LIST1: str
    LIST2: str
    MONTANT_HONORAIRES: float
    MONTANT_FRAIS: float
    NUMORDER: str
    PROCESS: str
    CL1: str
    CL2: str
    ID_CREATOR: int
    CCCREATOR: str
    CTCREATOR: str
    COMBIN: bytes
    COMBGCOLOR: float
    SALARYMIN: decimal.Decimal
    SALARYMAX: decimal.Decimal
    PERIOD: str
    ALLOTMENT: decimal.Decimal
    ANONYMOUS: int
    ID_CONSULTANT_3: int
    ZIP: str
    EDUCLEVEL: str
    JOBEXP: str
    TYPEJOB: str
    INVOICECONTACT: str
    ID_INVOICECONTACT: int
    SALARYGROUP: str
    CREATEDFACT1: int
    CREATEDFACT2: int
    CREATEDFACT3: int
    NUMFACT1: str
    NUMFACT2: str
    NUMFACT3: str
    PERCENTFACT1: int
    PERCENTFACT2: int
    PERCENTFACT3: int
    CONFIDENTIAL: int
    CONTRACTAMOUNT: decimal.Decimal
    DISTRIBUTIONWITHMARGE: int
    SHOWEMAILASSI: int
    URL: str
    POSTIT: str
    DATE_MAJ: str
    INDEXED: int
    CAC_FILTER: str
    CAC_PERS1_ACCESS: int
    CAC_PERS2_ACCESS: int
    CAC_PERS3_ACCESS: int
    CAC_PERS4_ACCESS: int
    CAC_ACCESS: int
    CAC_SHOW_DOC: str
    EASYRECRUECAMPAIGNID: int
    ASSESSFIRSTID: int
    ID_ISO_CURRENCY: str
    ID_DOCREF: int
    DESC_CUSTOMER: str
    ENTITE: str


class AdmenJob(Struct):
    ID_ANNONCE: int
    ID_SUPPORT: int
    ID_MISSION: int
    REFERENCE: str
    LIBELLE: str
    TEXTE_ANNONCE: str
    ETAT_ANNONCE: int
    COUT: float
    IMPUTABLE: int
    DATE_DEBUT: str
    DATE_FIN: str
    HTML: str
    SOFT_DELETED: int
    DATE_DELETED: str
    ID_PERDEL: int
    TARIF: str
    TYPE: str
    AGENCE: int
    FACTURE: str
    COLOR: int
    DESCRCUSTOMER: str
    DESCRASSIGNMENT: str
    ID_CREATOR: int
    CCCREATOR: str
    CTCREATOR: str
    VALIDATED: int
    METADESCRIPTION: str
    METAKEYS: str
    COMMENTAIRES: str
    DATE_CREATION: str
    ID_MULTIPOSTER: int
    STATUSMULTIPOSTER: int
    URL: str
    TEXTE_ANNONCE_TXT: str
    HTML_TXT: str
    DESCRCUSTOMER_TXT: str
    DESCRASSIGNMENT_TXT: str


class AdmenCandidature(Struct):
    ID_MISSION: int
    ID_PERSONNE: int
    PRESELECTIONNE: int
    SELECTIONNE: int
    CHOISI: int
    ENTRETIEN: int
    DATE_ENTRETIEN: str
    FIRST_ENTRETIEN: int
    DATE_1ER_ENTRETIEN: str
    CR_ENTRETIEN: str
    CR_RDV_1: str
    QUALIFIE: int
    ANNONCE: int
    EVALUATION: int
    SECOND_ENTRETIEN: int
    THIRD_ENTRETIEN: int
    OFFRE: int
    CR_RDV_2: str
    CR_RDV_3: str
    ACCUSE_RECEPT: int
    DATE_RDV_2: str
    DATE_RDV_3: str
    ID_ANNONCE: int
    DATE_AR: str
    INTERNE_ENTRETIEN: str
    EXTERNE_ENTRETIEN: str
    INTERNE_RDV1: str
    EXTERNE_RDV1: str
    INTERNE_RDV2: str
    EXTERNE_RDV2: str
    INTERNE_RDV3: str
    EXTERNE_RDV3: str
    GRAPHOLOGIE: int
    REPONSE_NEGATIVE: int
    date_reponse_neg: str
    NON_INTERESSE: int
    DESISTEMENT: int
    TYPE_DE_CONTRAT: str
    CAND_SAL_FIXE: int
    CAND_SAL_VARIABLE: int
    CAND_SALAIRE: int
    CAND_SAL_ANNEE: int
    CAND_STOCK_OPTION: int
    CAND_SAL_GROUPE: str
    CAND_SAL_AVANTAGE: str
    ORIGINE: str
    DUREE: str
    NON_INTERESSANT: int
    COLOR: str
    TYPE: int
    ID_ADVERTLINES: int
    ID_PERSONNE_WEB: int
    CHECK_C16: int
    CHECK_C17: int
    CHECK_C18: int
    CHECK_C19: int
    CHECK_C20: int
    CHECK_C21: int
    CHECK_C22: int
    CHECK_C23: int
    CHECK_C24: int
    CHECK_C25: int
    DATE_C1: str
    DATE_C2: str
    DATE_C3: str
    DATE_C4: str
    DATE_C5: str
    DATE_C6: str
    DATE_C7: str
    DATE_C8: str
    DATE_C9: str
    DATE_C10: str
    DATE_C11: str
    DATE_C12: str
    DATE_C13: str
    DATE_C14: str
    DATE_C15: str
    DATE_C16: str
    DATE_C17: str
    DATE_C18: str
    DATE_C19: str
    DATE_C20: str
    DATE_C21: str
    DATE_C22: str
    DATE_C23: str
    DATE_C24: str
    DATE_C25: str
    ID_ENVOI: int
    COMMENTS: str
    DATETOPERSON: str
    EXPORTED: int
    DATEEXPORT: str
    NAME: str
    FIRSTNAME: str
    MOBILEPHONE: str
    ID_CANDIDATURE: int
    ID_EASYRECRUECAND: int
    CUSTOMER_FEEDBACK: int
    APPLICATIONCOLOR: int
    Libre1: str


class AdmenProfessionalExperience(Struct):
    ID_PERSONNE: int
    DATE_EXP: datetime.date
    RAISON_SOCIALE: str
    NB_SUPLEANTS: int
    COMMENTAIRES: str
    TECHNIQUE: int
    POSTE_OCCUPE: str
    DATE_FIN: datetime.date
    SECTEUR: str
    EN_COURS: int
    CA_GERE: int
    ID_SOCIETE: int
    INTITULE_POSTE: str
    T_ID_ORGANIGRAMME: int
    REMUNERATION: int
    CURRENCY: str
    ID_BOSS: int
    ID_PERSONNE_WEB: int
    SECONDARYPOSTE: int
    DEPARTMENT: str
    ID_EXPERIENCE: int
    ID_COMPADDR: int


class AdmenProfile(Struct):
    ID_PERSONNE: int
    PER_ID_PERSONNE: int
    ID_SOCIETE: int
    CIVILITE: str
    NOM: str
    PRENOM: str
    ADRESSE: str
    CODE_POSTAL: str
    VILLE: str
    TEL_DIRECT: str
    TEL_STANDARD: str
    TEL_PERSO: str
    TEL_MOBILE: str
    EMAIL: str
    PAGER: str
    FAX_BUREAU: str
    FAX_PERSO: str
    POSTE_OCCUPE: str
    DATE_NAISSANCE: datetime.date
    CHASSE: int
    PROFIL_CANDIDAT: int
    DEJA_CANDIDAT: int
    DEJA_RENCONTRE: int
    CV: int
    DATE_CV: datetime.date
    MOBILITE: int
    EMAIL_PERSO: str
    MOBILITE_TXT: str
    ID_CONSULTANT: int
    GROUPE_SALAIRE: str
    SALAIRE_FIXE: int
    SALAIRE_VARIABLE: int
    SALAIRE_AVANTAGE: str
    ENCADREMENT: int
    NOMBRE_PERSONNES: int
    COMMENTAIRE: str
    PROSPECT: int
    CLIENT: int
    CANDIDAT: int
    TYPOLOGIE: int
    STOCK_OPTION: int
    EMAIL_DEFAUT: int
    MOBILE_DEFAULT: int
    SITU_FAMILLE: str
    LIBRE1: str
    LIBRE2: str
    LIBRE3: str
    LIBRE4: str
    TEL1_CONFIDENTIEL: int
    TEL2_CONFIDENTIEL: int
    TEL3_CONFIDENTIEL: int
    TEL4_CONFIDENTIEL: int
    GRAPHOLOGIE: int
    DATE_CREATION: datetime.date
    DATE_MAJ: datetime.datetime
    DATE_DEPUIS: datetime.date
    NATIONALITE: str
    SALAIRE: int
    LIB_SOCIETE: str
    NOM_SECRETAIRE: str
    PRENOM_SECRETAIRE: str
    TEL_SECRETAIRE: str
    REFERENCE: int
    INTITULE_POSTE: str
    POSTE_RECHERCHE: str
    INTERNET: int
    SPONTANEE: int
    PAYS: str
    VAL_NOTE1: int
    VAL_NOTE2: int
    VAL_NOTE3: int
    VAL_NOTE4: int
    FOURNISSEUR: int
    PRESCRIPTEUR: int
    PARTENAIRE: int
    ACTIF: int
    DOSSIER: int
    OFFLIMIT_AUTO: int
    OFFLIMIT_MANU: int
    PUBLIE: int
    ANNEE_SALAIRE: int
    ORIGINE: str
    DATE_MAJ_MANU: datetime.datetime
    TEL_PERSO2: str
    TEL_PERSO3: str
    MOBILE_PRO: str
    EN_COURS: int
    SOFT_DELETED: int
    DATE_DELETED: datetime.date
    ID_PERDEL: int
    COLOR: int
    LASTACTION: str
    CL1: str
    CL2: str
    CL3: str
    CS1: str
    CS2: str
    LIST1: str
    LIST2: str
    LIST3: str
    LIST4: str
    AREA: str
    CURRENCY: str
    MASK11: int
    MASK12: str
    MASK13: str
    MASK14: str
    MASK15: str
    COMPNAME: str
    ID_CREATOR: int
    CCCREATOR: str
    CTCREATOR: str
    SECTEUR: str
    COMBIN: str
    COMBGCOLOR: float
    ID_PERSONNE_WEB: int
    LASTDATEACTION: datetime.date
    LASTLIBACTION: str
    VALIDATED: int
    ASSISTANTMAIL: str
    LASTUSERACTION: str
    IMPORTED: int
    ID_LINKEDIN: str
    LINKEDINPUBLICPROFIL: str
    scsync_uri_viadeo: str
    scsync_uri_xing: str
    scsync_modified: str
    ID_AWA2: int
    POSTIT: str
    ID_VIADEO: str
    VIADEOPUBLICPROFIL: str
    INDEXED: int
    CAC_USER: str
    CAC_PWD: str
    CAC_DATE_END: datetime.date
    EXPECTEDSALARYMIN: int
    EXPECTEDSALARYMAX: int
    EMAILINGCONTACT: int
    SKYPELOGIN: str
    TWITTERPAGE: str
    FACEBOOKPAGE: str
    ID_COMPADDR: int
    ID_DOCREF: int
    GENDER: str
    ID_ISO_CURRENCY: str
    RGPD_VALIDATED: int
    RGPD_VALIDATEDDATE: datetime.date
    RGPD_ASKEDDATE: datetime.date
    RGPD_DISABLED: int
    Libre5: str
    Libre6: str
    Libre7: str
    Libre8: str
    ID_CONTACT: int



import typing as t

from hrflow_connectors.v2.connectors.workable.warehouse import WorkableWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_location(workable_job: t.Dict) -> t.Dict:
    location = workable_job.get("location", {})

    coords = next(
        (
            location["coords"].split(", ")
            for location in workable_job.get("locations", [])
            if location["city"] == location.get("city")
        ),
        None,
    )

    lat = float(coords[0]) if coords is not None else None
    lng = float(coords[1]) if coords is not None else None

    text = location.get("location_str", None)
    if not text:
        text_components = []
        for key in ["city", "region", "zip_code", "country"]:
            if location.get(key):
                text_components.append(location[key])
        text = ", ".join(text_components)

    geojson = dict(
        text=text,
        country=location.get("country"),
        state=location.get("region"),
        city=location.get("city"),
        postcode=location.get("zip_code"),
    )
    return dict(text=text, lat=lat, lng=lng, geojson=geojson)


def format_job(workable_job: t.Dict) -> t.Dict:
    """
    Format a job into the hrflow job object format
    Args:
        data (WorkableJobModel): a job object pulled from workable subdomain
    Returns:
        HrflowJob: a job into the hrflow job object format
    """
    t = lambda name, value: dict(name=name, value=value)

    hrflow_job = dict(
        name=workable_job.get("title"),
        reference=workable_job.get("shortcode"),
        url=workable_job.get("url"),
        location=get_location(workable_job),
        sections=[
            dict(
                name=field_name,
                title=field_name,
                description=workable_job.get(field_name),
            )
            for field_name in [
                "full_description",
                "description",
                "requirements",
                "benefits",
            ]
        ],
        created_at=workable_job.get("created_at"),
        summary=workable_job.get("description"),
        requirements=workable_job.get("requirements"),
        benefits=workable_job.get("benefits"),
        tags=[
            t("id", workable_job.get("id")),
            t("state", workable_job.get("state")),
            t("department", workable_job.get("department")),
            t("application_url", workable_job.get("application_url")),
            t("employment_type", workable_job.get("employment_type")),
            t("industry", workable_job.get("industry")),
            t("function", workable_job.get("function")),
            t("experience", workable_job.get("experience")),
            t("education", workable_job.get("education")),
        ],
        ranges_float=[
            dict(
                name="salary",
                value_min=workable_job.get("salary", {}).get("salary_from"),
                value_max=workable_job.get("salary", {}).get("salary_to"),
                unit=workable_job.get("salary", {}).get("salary_currency"),
            )
        ],
    )

    return hrflow_job


def format_job_archive(
    workable_job: t.Dict,
) -> t.Dict:
    return dict(
        reference=workable_job.get("shortcode"),
    )


def format_candidate(
    workable_candidate: t.Dict,
) -> t.Dict:
    t = lambda name, value: dict(name=name, value=value)

    hrflow_profile = dict(
        reference=workable_candidate.get("id"),
        info=dict(
            full_name=workable_candidate.get("name"),
            first_name=workable_candidate.get("firstname"),
            last_name=workable_candidate.get("lastname"),
            summary=workable_candidate.get("summary"),
            picture=workable_candidate.get("image_url"),
            email=workable_candidate.get("email"),
            phone=workable_candidate.get("phone"),
            location=get_location(workable_candidate),
        ),
        created_at=workable_candidate.get("created_at"),
        updated_at=workable_candidate.get("updated_at"),
        educations=[
            dict(
                school=education.get("school"),
                title=(
                    education.get("degree", "")
                    + " in "
                    + education.get("field_of_study")
                    if education.get("degree") and education.get("field_of_study")
                    else education.get("degree")
                    or education.get("field_of_study")
                    or "Undefined"
                ),
                description=None,
                date_start=education.get("start_date"),
                date_end=education.get("end_date"),
                location=dict(text=None, lat=None, lng=None),
            )
            for education in workable_candidate.get("education_entries", [])
        ],
        experiences=[
            dict(
                company=experience.get("company"),
                title=experience.get("title"),
                date_start=experience.get("start_date"),
                date_end=experience.get("end_date"),
                description=experience.get("summary"),
                location=dict(text=None, lat=None, lng=None),
            )
            for experience in workable_candidate.get("experience_entries", [])
        ],
        skills=[
            dict(name=skill["name"], value=None)
            for skill in workable_candidate.get("skills", [])
        ],
        tags=[
            t("job", workable_candidate.get("job")),
            t("stage", workable_candidate.get("stage")),
            t("disqualified", workable_candidate.get("disqualified")),
            t("disqualified_at", workable_candidate.get("disqualified_at")),
            t("disqualified_reason", workable_candidate.get("disqualified_reason")),
            t("hired_at", workable_candidate.get("hired_at")),
            t("source", workable_candidate.get("source")),
            t("answers", workable_candidate.get("answers")),
            t("workable_tags", workable_candidate.get("tags")),
        ],
        resume=dict(raw=workable_candidate.get("resume")),
    )

    return hrflow_profile


def format_candidate_archive(
    workable_candidate: t.Dict,
) -> t.Dict:
    return dict(
        reference=workable_candidate.get("id"),
    )


def format_profile(
    hrflow_profile: t.Dict,
) -> t.Dict:
    """
    Format a HrflowProfile object into a WorkableCandidate object
    Args:
        data (HrflowProfile): HrflowProfile object
    Returns:
        WorkableCandidate: WorkableCandidate object
    """
    info = hrflow_profile.get("info", {})

    resume_url = next(
        (
            attachment["public_url"]
            for attachment in hrflow_profile.get("attachments", [])
            if attachment["type"] == "resume"
        ),
        None,
    )

    candidate = dict(
        name=info.get("full_name"),
        firstname=info.get("first_name"),
        lastname=info.get("last_name"),
        email=info.get("email"),
        phone=info.get("phone"),
        address=info.get("location", {}).get("text"),
        summary=info.get("summary"),
        image_url=info.get("picture"),
        education_entries=[
            dict(
                school=education.get("school") or "Undefined",
                degree=education.get("title"),
                field_of_study=None,
                start_date=education.get("date_start"),
                end_date=education.get("date_end"),
            )
            for education in hrflow_profile.get("educations", [])
        ],
        experience_entries=[
            dict(
                company=experience.get("company"),
                title=experience.get("title") or "Undefined",
                summary=experience.get("description"),
                start_date=experience.get("date_start"),
                end_date=experience.get("date_end"),
            )
            for experience in hrflow_profile.get("experiences", [])
        ],
        skills=[skill["name"] for skill in hrflow_profile.get("skills", [])],
        social_profiles=[
            dict(
                type=url["type"],
                url=url["url"],
            )
            for url in info.get("urls", [])
        ],
    )

    if resume_url:
        candidate["resume_url"] = resume_url

    return candidate


def fromat_profile_update(
    hrflow_profile: t.Dict,
) -> t.Dict:
    """
    Format a HrflowProfile object into a WorkableCandidate object
    Args:
        data (HrflowProfile): HrflowProfile object
    Returns:
        WorkableCandidate: WorkableCandidate object
    """

    candidate = format_profile(hrflow_profile)
    candidate["id"] = hrflow_profile.get("reference")
    return candidate


Workable = Connector(
    name="Workable",
    type=ConnectorType.HCM,
    subtype="workable",
    warehouse=WorkableWarehouse,
    description=(
        "More than an applicant tracking system, "
        "Workable's talent acquisition software helps teams find candidates, "
        "evaluate applicants and make the right hire, faster."
    ),
    url="https://www.workable.com/",
    flows=(
        Flow(
            Mode.create,
            Entity.job,
            Direction.inbound,
            format=format_job,
        ),
        Flow(
            Mode.update,
            Entity.job,
            Direction.inbound,
            format=format_job,
        ),
        Flow(
            Mode.archive,
            Entity.job,
            Direction.inbound,
            format=format_job_archive,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_candidate,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_candidate,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_candidate_archive,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=fromat_profile_update,
        ),
    ),
)



import os
import time
from logging import LoggerAdapter
from typing import Optional, Tuple

from smbprotocol.connection import Connection
from smbprotocol.session import Session
from smbprotocol.tree import TreeConnect


def establish_smb_connection(
    adapter: LoggerAdapter,
    server: str,
    username: str,
    password: str,
    share_name: str,
    max_retries: int = 5,
    retry_delay: int = 5,
) -> Tuple[bool, Optional[TreeConnect], Optional[Session], Optional[Connection]]:
    """
    Establish an SMB connection to the specified share.

    Returns a tuple containing:
        - smb_connected (bool): Whether the connection was successful.
        - smb_tree (TreeConnect): SMB tree connection object.
        - smb_session (Session): SMB session object.
        - smb_connection (Connection): SMB connection object.
    """
    for retries in range(max_retries):
        try:
            # Establish a connection to the SMB server
            smb_connection = Connection(
                guid=os.urandom(16),
                server_name=server,
                port=445,
            )
            smb_connection.connect()

            # Create and authenticate a session
            smb_session = Session(
                smb_connection,
                username,
                password,
            )
            smb_session.connect()

            # Connect to the specified share
            smb_tree = TreeConnect(
                smb_session,
                f"\\\\{server}\\{share_name}",
            )
            smb_tree.connect()

            adapter.info("SMB connection established successfully.")
            return True, smb_tree, smb_session, smb_connection
        except Exception as e:
            adapter.error(f"SMB connection error: {e}")
            if retries < max_retries - 1:
                adapter.error(
                    f"Retrying SMB connection in {retry_delay} seconds... "
                    f"({retries + 1}/{max_retries})"
                )
                time.sleep(retry_delay)
            else:
                adapter.error("Max SMB connection retries reached.")
    return False, None, None, None



import json
import typing as t
from datetime import datetime, timedelta
from enum import Enum
from io import BytesIO
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.breezyhr.schemas import (
    BreezyJobModel,
    BreezyProfileModel,
)
from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)

BREEZY_BASE_URL = "https://api.breezy.hr/v3"


class State(str, Enum):
    draft = "draft"
    archived = "archived"
    published = "published"
    closed = "closed"
    pending = "pending"


class Origin(str, Enum):
    sourced = "sourced"
    applied = "applied"


class AuthParameters(Struct):
    email: Annotated[
        str,
        Meta(
            description="email",
        ),
    ]
    password: Annotated[str, Meta(description="password")]
    company_id: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "ID of company to pull jobs from in Breezy HR database associated with"
                " the authenticated user"
            ),
        ),
    ] = None
    company_name: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "[⚠️ Requiered if company_id is not specified], the company associated"
                " with the authenticated user"
            ),
        ),
    ] = None


class ReadJobsParameters(Struct):
    state: Annotated[
        State,
        Meta(
            description=(
                "Specify an optional position state filter. e.g. draft, archived,"
                " published, closed, pending\nDefaults to published"
            ),
        ),
    ] = State.published


class WriteProfilesParameters(Struct):
    position_id: Annotated[
        str,
        Meta(
            description="Id of the position to create a new candidate for",
        ),
    ]
    origin: Annotated[
        Origin,
        Meta(
            description=(
                "will indicate in Breezy if the candidate should be marked as sourced"
                " or applied"
            ),
        ),
    ] = Origin.sourced
    stage_actions_enabled: Annotated[
        bool,
        Meta(
            description=(
                'When origin is "sourced", should stage actions be executed (defaults'
                ' to false). This is always true when origin is "applied"'
            ),
        ),
    ] = False


class UpdateProfilesParameters(Struct):
    position_id: Annotated[
        str,
        Meta(
            description="Id of the position to create a new candidate for",
        ),
    ]


class ReadProfilesParameters(Struct):
    position_id: Annotated[
        t.Optional[str],
        Meta(
            description="Id of the position to create a new candidate for",
        ),
    ] = None


def get_access_token(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
):
    sign_in_url = f"{BREEZY_BASE_URL}/signin"

    response = requests.post(
        sign_in_url,
        json={"email": auth_parameters.email, "password": auth_parameters.password},
    )

    if not response.ok:
        adapter.error(f"failed to retrieve access token, reason: {response.text}")
        raise Exception("failed to retrieve access token")

    return response.json().get("access_token")


def revoke_access_token(access_token):
    sign_out_url = f"{BREEZY_BASE_URL}/signout"

    headers = {"Authorization": f"{access_token}"}

    requests.get(sign_out_url, headers=headers)


def generic_jobs_read(
    mode: Mode,
):
    def read_jobs(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadJobsParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        access_token = get_access_token(adapter, auth_parameters)
        headers = {"Authorization": access_token}

        company_id = auth_parameters.company_id
        if company_id is None:
            companies_url = f"{BREEZY_BASE_URL}/companies"

            companies_response = requests.get(companies_url, headers=headers)

            if not companies_response.ok:
                adapter.error(
                    f"Failed to retrieve company id, reason: {companies_response.text}"
                )
                raise Exception("Failed to retrieve company id")

            company_list = companies_response.json()
            for company in company_list:
                if company["name"] == auth_parameters.company_name:
                    company_id = company["_id"]
                    break
            if company_id is None:
                adapter.error(
                    "Failed to retrieve company id, reason: company does not match with"
                    " an id"
                )
                raise Exception("Failed to retrieve company id")

        url = f"{BREEZY_BASE_URL}/company/{company_id}/positions"
        params = {"state": parameters.state}

        response = requests.get(url, headers=headers, params=params)
        if not response.ok:
            adapter.error(f"Failed to read jobs, reason: {response.text}")
            raise Exception("Failed to read jobs")

        for job in response.json():
            if mode == Mode.create and not is_within_five_minutes(
                job["creation_date"], job["updated_date"]
            ):
                continue
            if mode == Mode.update and is_within_five_minutes(
                job["creation_date"], job["updated_date"]
            ):
                continue
            yield job

        revoke_access_token(access_token)

    return read_jobs


def send_profile(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    profiles: t.Iterable[t.Dict],
    access_token: str,
    company_id: str,
    candidate_id: str = "",
):
    base_url = (
        f"{BREEZY_BASE_URL}/company/{company_id}/"
        f"position/{parameters.position_id}/candidate"
    )

    payload = json.dumps(profiles)

    headers = {"Content-Type": "application/json", "Authorization": f"{access_token}"}

    if candidate_id == "":
        url = f"{base_url}s/"

        # If the candidate doesn't already exist we "POST" his profile
        response = requests.post(url, headers=headers, data=payload)

    else:
        # In case the candidate exists,
        #  we retrieve his id to update his profile with a "PUT" request

        url = f"{base_url}/{candidate_id}"

        url = (
            f"{BREEZY_BASE_URL}/company/{company_id}/"
            f"position/{parameters.position_id}/candidate/{candidate_id}"
        )

        adapter.info(f"Updating id = {candidate_id} profile")
        response = requests.put(url, headers=headers, data=payload)
    return response


def write_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    access_token = get_access_token(adapter, auth_parameters)
    headers = {"Authorization": access_token}

    company_id = auth_parameters.company_id

    if company_id is None:
        companies_url = f"{BREEZY_BASE_URL}/companies"

        companies_response = requests.get(companies_url, headers=headers)

        if not companies_response.ok:
            adapter.error(
                f"Failed to retrieve company id, reason: {companies_response.text}"
            )
            raise Exception("Failed to retrieve company id")

        company_list = companies_response.json()
        for company in company_list:
            if company["name"] == auth_parameters.company_name:
                company_id = company["_id"]
                break
        if company_id is None:
            adapter.error(
                "Failed to retrieve company id, reason: company does not match with"
                " an id"
            )
            raise Exception("Failed to retrieve company id")

    post_candidate_url = (
        f"{BREEZY_BASE_URL}/company/{company_id}/position/"
        f"{parameters.position_id}/candidates"
    )

    for profile in items:
        resume_url = profile.pop("resume", None)
        profile.update({"origin": parameters.origin})
        if parameters.stage_actions_enabled:
            params = {"stage_actions_enabled": parameters.stage_actions_enabled}
        else:
            params = {}

        post_candidate_response = requests.post(
            post_candidate_url, headers=headers, json=profile, params=params
        )
        if not post_candidate_response.ok:
            adapter.error(
                f"Couldn't create candidate, reason: {post_candidate_response.text}"
            )
            failed_profiles.append(profile)
            continue

        if resume_url:
            candidate_id = post_candidate_response.json()["_id"]
            resume_file_response = requests.get(url=resume_url)
            resume_file_obj = BytesIO(resume_file_response.content)
            files = {
                "file": (
                    "resume.pdf",
                    resume_file_obj,
                    "application/pdf",
                )
            }
            post_resume_url = (
                f"{BREEZY_BASE_URL}/company/{company_id}/position/"
                f"{parameters.position_id}/candidate/{candidate_id}/resume"
            )
            post_resume_response = requests.post(
                post_resume_url, headers=headers, files=files
            )
            if not post_resume_response.ok:
                adapter.error(
                    "Failed to attach resume to candidate, reason:"
                    f" {post_resume_response.text}"
                )

    revoke_access_token(access_token)
    return failed_profiles


def update_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    access_token = get_access_token(adapter, auth_parameters)
    headers = {"Authorization": access_token}

    company_id = auth_parameters.company_id
    if company_id is None:
        companies_url = f"{BREEZY_BASE_URL}/companies"

        companies_response = requests.get(companies_url, headers=headers)

        if not companies_response.ok:
            adapter.error(
                f"Failed to retrieve company id, reason: {companies_response.text}"
            )
            raise Exception("Failed to retrieve company id")

        company_list = companies_response.json()
        for company in company_list:
            if company["name"] == auth_parameters.company_name:
                company_id = company["_id"]
                break
        if company_id is None:
            adapter.error(
                "Failed to retrieve company id, reason: company does not match with"
                " an id"
            )
            raise Exception("Failed to retrieve company id")

    base_url = (
        f"{BREEZY_BASE_URL}/company/{company_id}/position/"
        f"{parameters.position_id}/candidate"
    )

    for profile in items:
        candidate_id = profile.pop("id")
        candidate_work_history = profile.pop("work_history", [])
        candidate_education = profile.pop("education", [])
        put_candidate_url = f"{base_url}/{candidate_id}"
        put_candidate_response = requests.put(
            put_candidate_url, headers=headers, json=profile
        )
        if not put_candidate_response.ok:
            adapter.error(
                f"Couldn't create candidate, reason: {put_candidate_response.text}"
            )
            failed_profiles.append(profile)
            continue

        current_educations = [
            {education["school_name"], education["field_of_study"]}
            for education in put_candidate_response.json().get("education", [])
        ]
        current_work_history = [
            {work_history["company_name"], work_history["title"]}
            for work_history in put_candidate_response.json().get("work_history", [])
        ]

        for education in candidate_education:
            if {
                education["school_name"],
                education["field_of_study"],
            } in current_educations:
                continue
            education_url = f"{put_candidate_url}/education"
            education_response = requests.put(
                education_url, headers=headers, json=education
            )
            if not education_response.ok:
                adapter.error(
                    f"Couldn't update education, for candidate {candidate_id}, reason:"
                    f" {education_response.text}"
                )
        for work_history in candidate_work_history:
            if {
                work_history["company_name"],
                work_history["title"],
            } in current_work_history:
                continue
            work_history_url = f"{put_candidate_url}/work-history"
            work_history_response = requests.put(
                work_history_url, headers=headers, json=work_history
            )
            if not work_history_response.ok:
                adapter.error(
                    f"Couldn't update experience, for candidate {candidate_id}, reason:"
                    f" {work_history_response.text}"
                )

    revoke_access_token(access_token)
    return failed_profiles


def generic_profiles_read(
    mode: Mode,
):
    def read_profiles(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadProfilesParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        # TODO: add incremental read_mode using page_size, and page as read_from
        access_token = get_access_token(adapter, auth_parameters)
        headers = {"Authorization": access_token}

        company_id = auth_parameters.company_id
        if company_id is None:
            companies_url = f"{BREEZY_BASE_URL}/companies"

            companies_response = requests.get(companies_url, headers=headers)

            if not companies_response.ok:
                adapter.error(
                    f"Failed to retrieve company id, reason: {companies_response.text}"
                )
                raise Exception("Failed to retrieve company id")

            company_list = companies_response.json()
            for company in company_list:
                if company["name"] == auth_parameters.company_name:
                    company_id = company["_id"]
                    break
            if company_id is None:
                adapter.error(
                    "Failed to retrieve company id, reason: company does not match with"
                    " an id"
                )
                raise Exception("Failed to retrieve company id")
        if not parameters.position_id:
            # retrieve all postion ids for all published positions
            positions_url = (
                f"{BREEZY_BASE_URL}/company/{company_id}/positions?state=published"
            )

            positions_response = requests.get(positions_url, headers=headers)
            if len(positions_response.json()) == 0:
                adapter.info("No published position found")
            for position in positions_response.json():
                position_id = position["_id"]
                candidates_url = (
                    f"{BREEZY_BASE_URL}/company/{company_id}/position/"
                    f"{position_id}/candidates?sort=updated"
                )
                candidates_response = requests.get(candidates_url, headers=headers)
                if not candidates_response.ok:
                    adapter.error(
                        "Failed to retrieve candidates, reason:"
                        f" {candidates_response.text}"
                    )
                    raise Exception("Failed to retrieve candidates")
                candidates = candidates_response.json()
                for candidate in candidates:
                    if mode == Mode.create and not is_within_five_minutes(
                        candidate["creation_date"], candidate["updated_date"]
                    ):
                        continue
                    if mode == Mode.update and is_within_five_minutes(
                        candidate["creation_date"], candidate["updated_date"]
                    ):
                        continue
                    candidate_id = candidate["_id"]
                    full_candidate_url = (
                        f"{BREEZY_BASE_URL}/company/{company_id}/position/"
                        f"{position_id}/candidate/{candidate_id}"
                    )
                    full_candidate_response = requests.get(
                        full_candidate_url, headers=headers
                    )
                    if not full_candidate_response.ok:
                        adapter.error(
                            "Failed to retrieve candidate, reason:"
                            f" {full_candidate_response.text}"
                        )
                        continue

                    full_candidate = full_candidate_response.json()

                    resume_url = full_candidate.get("resume", {}).get("url")
                    if not resume_url:
                        full_candidate["resume"] = None
                    if resume_url:
                        resume_response = requests.get(resume_url, headers=headers)
                        if not resume_response.ok:
                            adapter.error(
                                "Failed to retrieve resume, reason:"
                                f" {resume_response.text}"
                            )
                            full_candidate["resume"] = None
                        else:
                            full_candidate["resume"] = dict(raw=resume_response.content)

                    yield full_candidate
        else:
            candidates_url = (
                f"{BREEZY_BASE_URL}/company/{company_id}/position/"
                f"{parameters.position_id}/candidates?sort=updated"
            )

            candidates_response = requests.get(candidates_url, headers=headers)
            if not candidates_response.ok:
                adapter.error(
                    f"Failed to retrieve candidates, reason: {candidates_response.text}"
                )
                raise Exception("Failed to retrieve candidates")
            candidates = candidates_response.json()
            for candidate in candidates:
                if mode == Mode.create and not is_within_five_minutes(
                    candidate["creation_date"], candidate["updated_date"]
                ):
                    continue
                if mode == Mode.update and is_within_five_minutes(
                    candidate["creation_date"], candidate["updated_date"]
                ):
                    continue
                candidate_id = candidate["_id"]
                full_candidate_url = (
                    f"{BREEZY_BASE_URL}/company/{company_id}/position/"
                    f"{parameters.position_id}/candidate/{candidate_id}"
                )
                full_candidate_response = requests.get(
                    full_candidate_url, headers=headers
                )
                if not full_candidate_response.ok:
                    adapter.error(
                        "Failed to retrieve candidate, reason:"
                        f" {full_candidate_response.text}"
                    )
                    continue

                full_candidate = full_candidate_response.json()

                resume_url = full_candidate.get("resume", {}).get("url")
                if not resume_url:
                    full_candidate["resume"] = None
                if resume_url:
                    resume_response = requests.get(resume_url, headers=headers)
                    if not resume_response.ok:
                        adapter.error(
                            f"Failed to retrieve resume, reason: {resume_response.text}"
                        )
                        full_candidate["resume"] = None
                    else:
                        full_candidate["resume"] = dict(raw=resume_response.content)

                yield full_candidate
        revoke_access_token(access_token)

    return read_profiles


# To account for the time differnece between the start of object creation on the platform
# and its completion
def is_within_five_minutes(date_str1: str, date_str2: str) -> bool:
    date1 = datetime.fromisoformat(date_str1.replace("Z", "+00:00"))
    date2 = datetime.fromisoformat(date_str2.replace("Z", "+00:00"))

    difference = abs(date1 - date2)

    return difference <= timedelta(minutes=5)


JobsAisle = Aisle(
    name=Entity.job,
    schema=BreezyJobModel,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(
            create=generic_jobs_read(Mode.create),
            update=generic_jobs_read(Mode.update),
            archive=generic_jobs_read(Mode.archive),
        ),
    ),
)

ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=BreezyProfileModel,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=generic_profiles_read(Mode.create),
            update=generic_profiles_read(Mode.update),
            archive=generic_profiles_read(Mode.archive),
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters, update=UpdateProfilesParameters
        ),
        function=merge(create=write_profiles, update=update_profiles),
    ),
)
